/**
This file is part of a jTEM project.
All jTEM projects are licensed under the FreeBSD license 
or 2-clause BSD license (see http://www.opensource.org/licenses/bsd-license.php). 

Copyright (c) 2006-2010, Technische Universität Berlin, jTEM
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

-	Redistributions of source code must retain the above copyright notice, 
	this list of conditions and the following disclaimer.

-	Redistributions in binary form must reproduce the above copyright notice, 
	this list of conditions and the following disclaimer in the documentation 
	and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS 
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
OF SUCH DAMAGE.
**/

package de.jtem.java2d;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Paint;
import java.awt.Stroke;
import java.awt.geom.GeneralPath;
import java.awt.geom.Rectangle2D;
import java.util.List;


/** <p>A scene component whose geometry is a coordinate grid that fills a
 * rectangle in user space (usually the viewport of a viewer).</p>
 * 
 * <p>The spacing of
 * the grid lines is either 1, 2 or 5 times a power of ten choosen in such a
 * way that neither the number of horizontal grid lines nor the number of
 * vertical grid lines exceeds a limit set by the property
 * <code>maximumLineCount</code>.</p>
 * 
 * <p>Every ten'th grid line is overdrawn with a different type of line, which
 * form the "coarse grid". The paint and the stroke of these lines can be
 * controlled by editing the <code>SceneComponent</code> returned by
 * <code>getCoarseGrid</code>.</p>
 * */

public class CoordinateGrid extends SceneComponent {
	
	int maximumLineCount = 40;
	Rectangle2D.Double rectangle;
	SceneComponent coarseGrid;
	boolean keepingAspectRatio = true;
	final static double log10 = Math.log(10);
	String horizontalUnit = "";
	String verticalUnit = "";
	
	/** Creates an aspect ratio preserving grid filling the region where both
	 * coordinates lie between -1 and 1. */
	
	public CoordinateGrid() {
		this(new Rectangle2D.Double(-1,-1,2,2), true);
	}
	
	/** Creates a grid filling the rectangle <code>r</code>, preserving the
	 * aspect ratio if <code>keepAspectRatio</code> is true. */
	
	public CoordinateGrid(Rectangle2D.Double r, boolean keepAspectRatio) {
		rectangle = new Rectangle2D.Double(r.x,r.y,r.width,r.height);
		
		setFilled(Boolean.FALSE);
		setOutlined(Boolean.TRUE);
		setOutlinePaint(Color.white);
		setStroke(new BasicStroke());

		coarseGrid = new SceneComponent();
		coarseGrid.setFilled(Boolean.FALSE);
		coarseGrid.setOutlined(Boolean.TRUE);
		//coarseGrid.setOutlinePaint(Color.white);
		coarseGrid.setStroke(new BasicStroke(3));
		addChild(coarseGrid);
		
		update();
	}

	/** Returns the maximal number of horizontal lines both (as well as the
	 * maximal number of vertical lines). */
	
	public int getMaximumLineCount() {
		return maximumLineCount;
	}

	/** Sets the maximal number of horizontal lines both (as well as the
	 * maximal number of vertical lines). */
	
	public void setMaximumLineCount(int n) {
		if (n!= maximumLineCount) {
			maximumLineCount = n;
			update();
		}
	}
	
	/** Whether the horizontal spacing of grid lines equals the vertical
	 * spacing (in user space). */
	
	public boolean isKeepingAspectRatio() {
		return keepingAspectRatio;
	}
	
	/** Sets the boolean that determines whether the horizontal spacing of grid
	 * lines equals the vertical spacing (in user space). */
	
	public void setKeepingAspectRatio(boolean b) {
		keepingAspectRatio = b;
	}
	
	public String getHorizontalUnit() {
		return horizontalUnit;
	}
	
	public void setHorizontalUnit(String s) {
		horizontalUnit = s;
	}
	
	public String getVerticalUnit() {
		return verticalUnit;
	}
	
	public void setVerticalUnit(String s) {
		verticalUnit = s;
	}
	
	/** Returns the rectangle that is filled by the grid. */
	
	public Rectangle2D.Double getRectangle() {
		return (Rectangle2D.Double)rectangle.clone();
	}
	
	/** Sets the rectangle that is filled by the grid. */
	
	public void setRectangle(Rectangle2D.Double r) {
		if (!r.equals(rectangle)) {
			rectangle.setRect(r);
			update();
		}
	}
	
	/** Returns the scene component responsible for the coarse grid (every
	 * ten'th grid line). Edit its stroke and paint to control the appearance
	 * of the coarse grid. */
	
	public SceneComponent getCoarseGrid() {
		return coarseGrid;
	}
	
	void update() {
		GeneralPath path = new GeneralPath();
		double xSpacingMin = rectangle.width/maximumLineCount;
		double ySpacingMin = rectangle.height/maximumLineCount;
		if (keepingAspectRatio) {
			double spacingMin = Math.max(xSpacingMin,ySpacingMin);
			xSpacingMin = spacingMin;
			ySpacingMin = spacingMin;
		}
		double xPowerOf10 = Math.log(xSpacingMin)/log10;
		double xSpacing = Math.pow(10,Math.ceil(xPowerOf10));
		if(xSpacing>5*xSpacingMin) xSpacing /= 2.5;
		if(xSpacing>2*xSpacingMin) xSpacing /= 2;
		double xMajorSpacing = 10*xSpacing;

		double yPowerOf10 = Math.log(ySpacingMin)/log10;
		double ySpacing = Math.pow(10,Math.ceil(yPowerOf10));
		if(ySpacing>5*ySpacingMin) ySpacing /= 2.5;
		if(ySpacing>2*ySpacingMin) ySpacing /= 2;
		double yMajorSpacing = 10*ySpacing;
		
		double yMin = Math.floor(rectangle.y/ySpacing) * ySpacing;
		double yMax = rectangle.y+rectangle.height;

		for (double y = yMin; y<yMax; y+=ySpacing) {
		  path.moveTo( (float)rectangle.x, (float)y );
		  path.lineTo( (float)(rectangle.x+rectangle.width), (float)y );
		}

		double xMin = Math.floor(rectangle.x/xSpacing) * xSpacing;
		double xMax = rectangle.x+rectangle.width;

		for (double x = xMin; x<xMax; x+=xSpacing) {
		  path.moveTo( (float)x, (float)rectangle.y );
		  path.lineTo( (float)x, (float)(rectangle.y+rectangle.height));
		}
		
		setShape(path);
		path = new GeneralPath();

		int iMin = (int)Math.round(
			rectangle.y/yMajorSpacing + .5
		);
		int iMax = (int)Math.round(
			(rectangle.y+rectangle.height)/yMajorSpacing - .5
		);

		int jMin = (int)Math.round(
			rectangle.x/xMajorSpacing + .5
		);
		int jMax = (int)Math.round(
			(rectangle.x+rectangle.width)/xMajorSpacing - .5
		);
		
		List<Annotation> annotations = coarseGrid.getAnnotations();
		annotations.clear();
		for (int i = iMin; i<=iMax; i++) {
		  double y = i * yMajorSpacing;
		  path.moveTo( (float)rectangle.x, (float)y );
		  path.lineTo( (float)(rectangle.x+rectangle.width), (float)y );
		  annotations.add(
				  new Annotation(
						  format(y)+verticalUnit,
						  rectangle.x,
						  y,
						  Annotation.SOUTHWEST
				  )
		  );
		}

		for (int j = jMin; j<=jMax; j++) {
		  double x = j * xMajorSpacing;
		  path.moveTo( (float)x, (float)rectangle.y );
		  path.lineTo( (float)x, (float)(rectangle.y+rectangle.height));
		  annotations.add(
				  new Annotation(
						  format(x)+horizontalUnit,
						  x,
						  rectangle.y,
						  Annotation.SOUTHWEST
				  )
		  );
		}
		coarseGrid.setShape(path);
	}
	
	public Stroke getCoarseGridStroke() {
		return coarseGrid.getStroke();
	}
	
	public void setCoarseGridStroke(Stroke s) {
		coarseGrid.setStroke(s);
	}
	
	public Paint getCoarseGridPaint() {
		return coarseGrid.getPaint();
	}
	
	public void setCoarseGridPaint(Paint s) {
		coarseGrid.setPaint(s);
	}
	
	private String format(double x) {
		String num = String.valueOf(x);
		String[] split = num.split("E");
		String firstPart = split[0];
		int end = Math.min(6,firstPart.length());
		String shorter = firstPart.substring(0,end);
		return split.length == 1 ? shorter : shorter+"E"+split[1];
	}
}
